# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import InitVar
from typing import Any

from pydantic import EmailStr, Field, HttpUrl
from pydantic.dataclasses import dataclass
from requests import Response
from requests.exceptions import ConnectionError

from ikigai.client.api import AccessAPI, ComponentAPI, SearchAPI
from ikigai.client.session import Session, SSLConfig
from ikigai.utils.compatibility import HTTPMethod

logger = logging.getLogger("ikigai.client")


@dataclass
class Client:
    # Init only vars
    user_email: InitVar[EmailStr]
    api_key: InitVar[str]
    base_url: InitVar[HttpUrl]
    ssl: InitVar[SSLConfig]

    __session: Session = Field(init=False)
    __access_api: AccessAPI = Field(init=False)
    __component_api: ComponentAPI = Field(init=False)
    __search_api: SearchAPI = Field(init=False)

    def __post_init__(
        self, user_email: EmailStr, api_key: str, base_url: HttpUrl, ssl: SSLConfig
    ) -> None:
        self.__session = Session(
            user_email=user_email, api_key=api_key, base_url=base_url, ssl=ssl
        )
        self.__access_api = AccessAPI(session=self.__session)
        self.__component_api = ComponentAPI(session=self.__session)
        self.__search_api = SearchAPI(session=self.__session)

        # Validate Base URL
        self.__validate_base_url__()

    def __validate_base_url__(self) -> None:
        heartbeats = {"Search": self.__search_api.heartbeat}
        # Validate Base URL by making test requests
        service: str = ""
        try:
            for service, heartbeat in heartbeats.items():
                heartbeat()
        except ConnectionError:
            message = (
                f"Could not find Ikigai {service}API via "
                f"{self.__session.base_url}. Please check the base_url, your "
                "internet connection, and SSL configuration."
            )
            raise ValueError(message) from None
        except RuntimeError:
            message = (
                f"Failed to connect to Ikigai {service}API via "
                f"{self.__session.base_url}. Please check the base_url. If the "
                "problem persists, please report an issue."
            )
            raise ValueError(message) from None

    def get(self, path: str, params: dict[str, Any] | None = None) -> Response:
        return self.__session.request(method=HTTPMethod.GET, path=path, params=params)

    def post(self, path: str, json: dict[Any, Any] | None = None) -> Response:
        return self.__session.request(method=HTTPMethod.POST, path=path, json=json)

    # APIs

    @property
    def access(self) -> AccessAPI:
        return self.__access_api

    @property
    def component(self) -> ComponentAPI:
        return self.__component_api

    @property
    def search(self) -> SearchAPI:
        return self.__search_api



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, Literal, TypedDict

from ikigai.typing import DirectoryDict
from ikigai.utils import DatasetDownloadStatus
from ikigai.utils.compatibility import NotRequired

# -------------------------------------------------------------------------------------
# App Related Data Exchange Types


class AppDict(TypedDict):
    project_id: str
    name: str
    owner: str
    description: str
    icon: str
    images: list[str]
    directory: DirectoryDict
    created_at: str
    modified_at: str
    last_used_at: str


class GetComponentsForProjectResponse(TypedDict):
    charts: list[Mapping[str, Any]]
    connectors: list[Mapping[str, Any]]
    dashboards: list[Mapping[str, Any]]
    datasets: list[Mapping[str, Any]]
    databases: list[Mapping[str, Any]]
    pipelines: list[Mapping[str, Any]]
    models: list[Mapping[str, Any]]
    external_resources: list[Mapping[str, Any]]
    users: list[Mapping[str, Any]]
    connector_directories: list[Mapping[str, Any]]
    dashboard_directories: list[Mapping[str, Any]]
    dataset_directories: list[Mapping[str, Any]]
    database_directories: list[Mapping[str, Any]]
    pipeline_directories: list[Mapping[str, Any]]
    model_directories: list[Mapping[str, Any]]
    external_resource_directories: list[Mapping[str, Any]]


# -------------------------------------------------------------------------------------
# Custom Facet Related Data Exchange Types


class CustomFacetDict(TypedDict):
    custom_facet_id: str
    name: str
    chain_group: str
    facet_uid: str
    description: str
    tags: list[str]
    python_script: str
    libraries: list[str]
    rootkit_token: str
    arguments: list[CustomFacetArgumentDict]
    created_at: str
    modified_at: str


class CustomFacetArgumentDict(TypedDict):
    name: str
    type: Literal["bool", "int", "str", "float"]
    value: bool | int | str | float


class CustomFacetVersionDict(TypedDict):
    version: str
    version_id: str
    custom_facet_id: str
    description: str
    arguments: list[CustomFacetArgumentDict]
    created_at: str


# -------------------------------------------------------------------------------------
# Dataset Related Data Exchange Types


class GetDatasetMultipartUploadUrlsResponse(TypedDict):
    upload_id: str
    content_type: str
    urls: dict[int, str]


class _InitializeDatasetDownloadFailedResponse(TypedDict):
    status: Literal[DatasetDownloadStatus.FAILED]


class _InitializeDatasetDownloadInProgressResponse(TypedDict):
    status: Literal[DatasetDownloadStatus.IN_PROGRESS]


class _InitializeDatasetDownloadSuccessResponse(TypedDict):
    status: Literal[DatasetDownloadStatus.SUCCESS]
    url: str


InitializeDatasetDownloadResponse = (
    _InitializeDatasetDownloadFailedResponse
    | _InitializeDatasetDownloadInProgressResponse
    | _InitializeDatasetDownloadSuccessResponse
)


class DatasetDict(TypedDict):
    project_id: str
    dataset_id: str
    name: str
    filename: str
    data_types: dict[str, DataTypeDict]
    directory: DirectoryDict
    is_optimized: bool
    file_extension: str
    size: int
    is_visible: bool
    created_at: str
    modified_at: str


class DataTypeDict(TypedDict):
    data_type: str
    data_formats: str


class DatasetLogDict(TypedDict):
    status: str
    timestamp: str
    job_type: str


# -------------------------------------------------------------------------------------
# Directory Related Data Exchange Types

# -------------------------------------------------------------------------------------
# Flow Related Data Exchange Types

RunVariablesRequest = dict[str, dict[Literal["value"], Any]]


class ScheduleDict(TypedDict):
    name: str
    start_time: str
    end_time: NotRequired[str]
    cron: str


class FlowDict(TypedDict):
    project_id: str
    pipeline_id: str
    name: str
    directory: DirectoryDict
    definition: FlowDefinitionDict
    trigger_downstream_pipelines: bool
    high_volume_preference: bool
    schedule: NotRequired[ScheduleDict]
    last_run: NotRequired[dict]
    next_run: NotRequired[dict]
    created_at: str
    modified_at: str


class FlowDefinitionDict(TypedDict):
    facets: list[FacetDict]
    arrows: list[ArrowDict]
    variables: NotRequired[dict[str, FlowVariableDict]]
    model_variables: NotRequired[dict[str, FlowModelVariableDict]]


class FacetDict(TypedDict):
    facet_id: str
    facet_uid: str
    name: NotRequired[str]
    arguments: NotRequired[dict]


class ArrowDict(TypedDict):
    source: str
    destination: str
    arguments: NotRequired[dict]


class FlowVariableDict(TypedDict):
    name: str
    value: Any
    facet_name: NotRequired[str]
    type: str
    is_list: bool


class FlowModelVariableDict(TypedDict):
    facet_name: str
    model_name: str
    model_version: NotRequired[str]
    model_argument_type: str
    model_arguments: list[dict]


class FlowStatusReportDict(TypedDict):
    status: str
    progress: NotRequired[int]
    message: str


class FlowLogDict(TypedDict):
    log_id: str
    status: str
    user: str
    erroneous_facet_id: NotRequired[str]
    message: str
    timestamp: str


# -------------------------------------------------------------------------------------
# Model Related Data Exchange Types


class ModelDict(TypedDict):
    project_id: str
    model_id: str
    name: str
    latest_version_id: str
    directory: DirectoryDict
    model_type: str
    sub_model_type: str
    description: str
    created_at: str
    modified_at: str


class ModelVersionDict(TypedDict):
    version_id: str
    model_id: str
    version: str
    hyperparameters: dict
    metrics: dict
    created_at: str
    modified_at: str


HyperParameterName = str
HyperParameterGroupName = str
ModelHyperParameterGroupDict = dict[HyperParameterName, Any]
ModelHyperParameterGroupType = dict[
    HyperParameterGroupName, ModelHyperParameterGroupDict
]


# -------------------------------------------------------------------------------------
# Specs Related Data Exchange Types


class FacetSpecsDict(TypedDict):
    INPUT: dict[str, dict[str, FacetSpecDict]]
    MID: dict[str, dict[str, FacetSpecDict]]
    OUTPUT: dict[str, dict[str, FacetSpecDict]]


class FacetSpecDict(TypedDict):
    facet_info: FacetInfoDict
    is_deprecated: bool
    is_hidden: bool
    facet_keywords: list[str]
    facet_requirements: list[FacetRequirementDict]
    facet_arguments: list[FacetArgumentSpecDict]
    in_arrow_arguments: list[FacetArrowArgumentSpecDict]
    out_arrow_arguments: list[FacetArrowArgumentSpecDict]


class FacetInfoDict(TypedDict):
    chain_group: str
    facet_group: str
    facet_type: str
    facet_uid: str


class FacetRequirementDict(TypedDict):
    max_child_count: int
    max_parent_count: int
    min_child_count: int
    min_parent_count: int


class FacetArgumentSpecDict(TypedDict):
    name: str
    argument_type: str
    is_required: bool
    default_value: Any | None
    options: NotRequired[list[Any]]
    is_list: bool
    is_deprecated: bool
    is_hidden: bool
    have_sub_arguments: bool
    children: Mapping[str, FacetArgumentSpecDict]


class FacetArrowArgumentSpecDict(TypedDict):
    name: str
    argument_type: str
    is_required: bool
    options: NotRequired[list[Any]]
    is_list: bool
    is_deprecated: bool
    is_hidden: bool
    have_sub_arguments: bool
    children: Mapping[str, FacetArrowArgumentSpecDict]


class ModelSpecDict(TypedDict):
    name: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    sub_model_types: list[SubModelSpecDict]


class SubModelSpecDict(TypedDict):
    name: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    metrics: ModelMetricsSpecDict
    parameters: dict[str, ModelParameterSpecDict]
    hyperparameters: dict[str, ModelHyperparameterSpecDict]


class EmptyDict(TypedDict):
    pass


ModelMetricsSpecDict = dict[str, EmptyDict]


class ModelParameterSpecDict(TypedDict):
    name: str
    default_value: Any
    have_options: bool
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    options: list[Any]
    parameter_type: str


class ModelHyperparameterSpecDict(TypedDict):
    name: str
    default_value: Any
    have_options: bool
    have_sub_hyperparameters: bool
    hyperparameter_group: str | None
    hyperparameter_type: str
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    children: dict[str, ModelHyperparameterSpecDict]
    options: list[Any]
    sub_hyperparameter_requirements: list[tuple[Any, list[str]]]


# -------------------------------------------------------------------------------------



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

import math
import random
from collections.abc import Callable
from typing import Any

import pandas as pd
import pytest
from faker import Faker

from ikigai import Ikigai

_ColumnNames = list[str]
_ColumnGenerator = Callable[[], Any]
_ColumnGenerators = list[_ColumnGenerator]
_TableGenerator = tuple[_ColumnNames, _ColumnGenerators]


def _create_table_generator(fake: Faker, k: int) -> _TableGenerator:
    """
    Return a table generator that generates a table with randomly selected columns
    where, k <= number of columns <= 2k

    Parameters
    ----------
    fake: Faker
        The Faker instance to use for generating fake data

    k: int
        The number of column generator selections to make

    Returns
    -------
    _TableGenerator
        The table generator consisting of column names and their corresponding
        generator functions
    """
    available_column_generators: list[list[tuple[str, _ColumnGenerator]]] = [
        [("name", fake.name), ("age", lambda: random.randint(20, 90))],
        [("ssn", fake.ssn)],
        [
            ("lat", lambda: float(fake.latitude())),
            ("lon", lambda: float(fake.longitude())),
        ],
        [("date", lambda: str(fake.date_time()))],
        [("email", fake.free_email)],
        [("work_email", fake.company_email)],
        [("job", fake.job), ("salary", fake.pricetag)],
    ]

    column_generators_selection = random.choices(available_column_generators, k=k)

    table_column_names: _ColumnNames = []
    table_column_generators: _ColumnGenerators = []
    for idx, column_generators in enumerate(column_generators_selection, start=1):
        for column_name, column_generator in column_generators:
            table_column_names.append(f"{column_name}-{idx}")
            table_column_generators.append(column_generator)
    return table_column_names, table_column_generators


@pytest.fixture()
def ikigai(cred: dict[str, Any]) -> Ikigai:
    return Ikigai(**cred)


@pytest.fixture()
def app_name(random_name: str) -> str:
    return f"proj-{random_name}"


@pytest.fixture()
def app_directory_name(random_name: str) -> str:
    return f"app-dir-{random_name}"


def _generate_df(table_generator: _TableGenerator, num_rows: int) -> pd.DataFrame:
    column_names, column_generators = table_generator
    rows = [[gen() for gen in column_generators] for _ in range(num_rows)]
    return pd.DataFrame(data=rows, columns=column_names)


@pytest.fixture()
def df1(faker: Faker) -> pd.DataFrame:
    num_generator_selections = random.randint(1, 10)
    num_rows = math.ceil(random.triangular(1, 100))
    table_generator = _create_table_generator(faker, num_generator_selections)
    return _generate_df(table_generator=table_generator, num_rows=num_rows)


@pytest.fixture()
def df2(faker: Faker) -> pd.DataFrame:
    num_generator_selections = random.randint(1, 10)
    num_rows = math.ceil(random.triangular(1, 100))
    table_generator = _create_table_generator(faker, num_generator_selections)
    return _generate_df(table_generator=table_generator, num_rows=num_rows)


@pytest.fixture()
def df_ml_regression1(faker: Faker) -> pd.DataFrame:
    """
    Generate a DataFrame suitable for ML regression tasks

    Parameters
    ----------
    faker: Faker
        The Faker instance to use for generating fake data

    Returns
    -------
    pd.DataFrame
        The generated DataFrame suitable for regression tasks.
        The DataFrame contains a target column with continuous values drawn
        from a uniform distribution between 0 and 100.
    """
    num_generator_selections = random.randint(3, 10)
    num_rows = math.ceil(random.triangular(100, 1000))
    table_generator = _create_table_generator(faker, num_generator_selections)
    # Add a target column for regression
    target_column_name = "target"
    generated_df = _generate_df(table_generator=table_generator, num_rows=num_rows)
    generated_df[target_column_name] = generated_df.apply(
        lambda _: random.uniform(0, 100), axis=1
    )
    return generated_df


@pytest.fixture()
def df_ml_classification1(faker: Faker) -> pd.DataFrame:
    """
    Generate a DataFrame suitable for ML classification tasks

    Parameters
    ----------
    faker: Faker
        The Faker instance to use for generating fake data

    Returns
    -------
    pd.DataFrame
        The generated DataFrame suitable for classification tasks.
        The DataFrame contains a target column with categorical class labels,
        where the number of classes is randomly chosen between 2 and (num_rows / 3).
    """
    num_generator_selections = random.randint(3, 10)
    num_rows = math.ceil(random.triangular(100, 1000))
    num_classes = math.floor(random.triangular(2, num_rows / 3))
    table_generator = _create_table_generator(faker, num_generator_selections)
    # Add a target column for classification
    target_column_name = "target"
    generated_df = _generate_df(table_generator=table_generator, num_rows=num_rows)
    generated_df[target_column_name] = generated_df.apply(
        lambda _: random.choice([f"class-{i}" for i in range(1, num_classes + 1)]),
        axis=1,
    )
    return generated_df


@pytest.fixture()
def custom_facet_name(random_name: str) -> str:
    return f"custom-facet-{random_name}"


@pytest.fixture()
def custom_facet_name_2(random_name_2: str) -> str:
    return f"custom-facet-{random_name_2}"


@pytest.fixture()
def custom_facet_name_3(random_name_3: str) -> str:
    return f"custom-facet-{random_name_3}"


@pytest.fixture()
def custom_facet_version_name_1(random_name: str) -> str:
    return f"version-{random_name}"


@pytest.fixture()
def custom_facet_version_name_2(random_name_2: str) -> str:
    return f"version-{random_name_2}"


@pytest.fixture()
def custom_facet_version_name_3(random_name_3: str) -> str:
    return f"version-{random_name_3}"


@pytest.fixture()
def dataset_name(random_name: str) -> str:
    return f"dats-{random_name}"


@pytest.fixture()
def dataset_directory_name_1(random_name: str) -> str:
    return f"dats-dir-{random_name}"


@pytest.fixture()
def dataset_directory_name_2(random_name_2: str) -> str:
    return f"dats-dir-{random_name_2}"


@pytest.fixture()
def flow_name(random_name: str) -> str:
    return f"flow-{random_name}"


@pytest.fixture()
def flow_directory_name_1(random_name: str) -> str:
    return f"flow-dir-{random_name}"


@pytest.fixture()
def flow_directory_name_2(random_name_2: str) -> str:
    return f"flow-dir-{random_name_2}"


@pytest.fixture()
def model_name(random_name: str) -> str:
    return f"model-{random_name}"



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import ast
import textwrap
from collections.abc import Mapping
from datetime import datetime
from functools import cached_property
from logging import getLogger
from pathlib import Path
from typing import Any

from pydantic import (
    AliasChoices,
    BaseModel,
    EmailStr,
    Field,
    PrivateAttr,
    field_validator,
)

from ikigai.client import Client, datax
from ikigai.specs import (
    CustomFacetArgumentSpec,
    CustomFacetType,
    FacetType,
    FacetTypes,
)
from ikigai.typing import ComponentBrowser, NamedMapping
from ikigai.utils import CustomFacetAccessLevel, CustomFacetArgumentType
from ikigai.utils.compatibility import Self, deprecated

logger = getLogger("ikigai.components")


class CustomFacetBuilder:
    _name: str
    _facet_type: FacetType | None
    _description: str
    _script: str
    _requirements: list[str]
    _system_access: bool
    _arguments: dict[str, CustomFacetArgumentSpec]
    __client: Client

    def __init__(self, client: Client) -> None:
        """
        Initialize the CustomFacetBuilder

        Parameters
        ----------
        client : Client
            The Ikigai client to use for API interactions

        Returns
        -------
        None
        """
        self.__client = client
        self._name = ""
        self._facet_type = None
        self._description = ""
        self._script = ""
        self._requirements = []
        self._system_access = False
        self._arguments = {}

    def new(self, name: str, facet_type: FacetType) -> Self:
        self._name = name
        self._facet_type = facet_type
        return self

    def description(self, description: str) -> Self:
        """
        Specify the description for the custom facet.

        Parameters
        ----------
        description : str
            The description for the custom facet.

        Returns
        -------
        Self
            The CustomFacetBuilder instance with the description set.
        """
        self._description = description
        return self

    def script(
        self,
        script: str | Path,
        *,
        requirements: list[str] | Path | None = None,
        system_access: bool = False,
        arguments: dict[str, str | int | float | bool] | None = None,
    ) -> Self:
        """
        Set the script and related dependencies for the custom facet

        You can specify the script and any PyPI libraries,
        as well as the arguments that are required to run the script.
        By default, the script has limited access to filesystem and environment,
        but you can request for more access to the environment, filesystem, and syscalls
        by setting the system_access flag to True.

        Note
        -----
        The user must have rootkit access to request system access.

        Parameters
        ----------
        script : str | Path
            The draft script for the custom facet.
            Can either be the text of the script or
            a path to a file containing the script.

        requirements : list[str] | Path | None
            The requirements for the custom facet.
            Can either be a path to a file containing the requirements in pip format,
            or be a list of strings each representing a single requirement.

        system_access : bool, optional
            Whether to grant system access to the custom facet, by default False.
            If True, the custom facet will have unrestricted access to the
            python environment and libraries.
            Creating a custom facet with system access will fail if
            the user does not have rootkit access.

        arguments : dict[str, str | int | float | bool]
            The arguments for the custom facet script and their default values.

        Examples
        --------
        >>> script_file = Path("./script.py")
        >>> builder = ikigai.custom_facet.new(...)
        >>> builder.script(script=script_file)

        >>> builder = ikigai.custom_facet.new(...)
        >>> builder.script(
        ...    script="result = data  # no-op",
        ...    requirements=["scikit-learn", "pandas==1.5.2"],
        ...    system_access=True,
        ...    arguments={"input-1": "input-1"},
        ... )

        >>> script_file = Path("./script.py")
        >>> builder = ikigai.custom_facet.new(...)
        >>> builder.script(
        ...    script=script_file,
        ...    requirements=Path("requirements.txt"),
        ...    system_access=True,
        ...    arguments={"input-1": "input-1", "input-2": "input-2"},
        ... )

        Returns
        -------
        Self
            The CustomFacetBuilder instance with the script set.
        """
        if isinstance(script, Path):
            with script.open("r") as fp:
                script = fp.read()
        script = textwrap.dedent(script)
        # Validate the script syntax againt ikigai platform's python version
        # IPLT-11330: See if we can avoid hardcoding the python version
        ast.parse(script, feature_version=(3, 10))
        self._script = script

        if requirements is None:
            requirements = []
        if isinstance(requirements, Path):
            with requirements.open("r") as fp:
                requirements = fp.readlines()
        self._requirements = requirements

        if arguments is None:
            arguments = {}
        self._arguments = {
            name: CustomFacetArgumentSpec(
                name=name,
                argument_type=CustomFacetArgumentType.from_value(value),
                value=value,
            )
            for name, value in arguments.items()
        }

        self._system_access = system_access
        return self

    def build(self) -> CustomFacet:
        """
        Build the custom facet object.

        Creates the custom facet in the Ikigai platform using the provided
        parameters and returns the corresponding CustomFacet object.

        Returns
        -------
        CustomFacet
            The created CustomFacet object.
        """
        # Validate the facet type
        if self._facet_type is None:
            error_msg = "Facet type is required to build a CustomFacet"
            raise ValueError(error_msg)

        # If rootkit is required, generate it from the script
        rootkit_token = (
            self.__client.access.generate_rootkit_token(script=self._script)
            if self._system_access
            else ""
        )

        # Create the custom facet
        arguments = [argument.to_dict() for argument in self._arguments.values()]

        custom_facet_id = self.__client.component.create_custom_facet(
            name=self._name,
            chain_group=self._facet_type.facet_info.chain_group,
            description=self._description,
            tags=[],
            python_script=self._script,
            libraries=self._requirements,
            rootkit_token=rootkit_token,
            arguments=arguments,
        )

        custom_facet_dict = self.__client.component.get_custom_facet(
            custom_facet_id=custom_facet_id,
        )
        return CustomFacet.from_dict(data=custom_facet_dict, client=self.__client)


class CustomFacetAccess(BaseModel):
    """
    Access manager for the Custom Facet.

    Grant/Update/Revoke access to the Custom Facet.
    """

    __custom_facet_id: str = PrivateAttr()
    __client: Client = PrivateAttr()

    def __init__(self, *, custom_facet_id: str, client: Client) -> None:
        super().__init__()
        self.__custom_facet_id = custom_facet_id
        self.__client = client

    def grant(self, email: EmailStr, access_level: CustomFacetAccessLevel) -> Self:
        """
        Grant access to the Custom Facet.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to grant access to.

        access_level: CustomFacetAccessLevel
            Access level to grant to the user.

        Returns
        -------
        Self
            Access manager for the Custom Facet.
        """
        self.__client.component.grant_custom_facet_access(
            custom_facet_id=self.__custom_facet_id,
            email=email,
            access_level=access_level,
        )
        return self

    def update(self, email: EmailStr, access_level: CustomFacetAccessLevel) -> Self:
        """
        Update access to the Custom Facet.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to update access for.

        access_level: CustomFacetAccessLevel
            New access level for the user.

        Returns
        -------
        Self
            Access manager for the Custom Facet.
        """
        self.__client.component.update_custom_facet_access(
            custom_facet_id=self.__custom_facet_id,
            email=email,
            access_level=access_level,
        )
        return self

    def revoke(self, email: EmailStr) -> Self:
        """
        Revoke access to the Custom Facet.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to revoke access from.

        Returns
        -------
        Self
            Access manager for the Custom Facet.
        """
        self.__client.component.revoke_custom_facet_access(
            custom_facet_id=self.__custom_facet_id,
            email=email,
        )
        return self


class CustomFacet(BaseModel):
    custom_facet_id: str
    name: str
    facet_type: FacetType
    description: str
    script: str = Field(validation_alias=AliasChoices("script", "python_script"))
    requirements: list[str] = Field(
        validation_alias=AliasChoices("requirements", "libraries")
    )
    rootkit_token: str
    arguments: dict[str, CustomFacetArgumentSpec]
    created_at: datetime
    modified_at: datetime
    __client: Client = PrivateAttr()

    @field_validator("arguments", mode="before")
    @classmethod
    def validate_arguments(cls, v: list[dict]) -> dict[str, CustomFacetArgumentSpec]:
        return {
            argument["name"]: CustomFacetArgumentSpec.model_validate(argument)
            for argument in v
        }

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        facet_types = FacetTypes.from_dict(client.component.get_facet_specs())
        facet_type = facet_types.find_by_uid(uid=data["facet_uid"])
        self = cls.model_validate({**data, "facet_type": facet_type})
        self.__client = client
        return self

    """
    Operations on Custom Facet
    """

    def rename(self, name: str) -> Self:
        """
        Rename the custom facet.

        Parameters
        ----------
        name : str
            The new name for the custom facet.

        Returns
        -------
        Self
            The CustomFacet instance with the name set.
        """
        self.__client.component.edit_custom_facet(
            custom_facet_id=self.custom_facet_id,
            chain_group=self.facet_type.facet_info.chain_group,
            name=name,
        )
        self.name = name
        return self

    def update_description(self, description: str) -> Self:
        """
        Update the description for the custom facet.

        Parameters
        ----------
        description : str
            The new description for the custom facet.

        Returns
        -------
        Self
            The CustomFacet instance with the new description.
        """
        self.__client.component.edit_custom_facet(
            custom_facet_id=self.custom_facet_id,
            chain_group=self.facet_type.facet_info.chain_group,
            description=description,
        )
        self.description = description
        return self

    def update_script(
        self,
        script: str | Path,
        *,
        requirements: list[str] | Path | None = None,
        system_access: bool = False,
        arguments: dict[str, str | int | float | bool] | None = None,
    ) -> Self:
        """
        Update the script and related dependencies for the custom facet

        You can specify the script and any PyPI libraries,
        as well as the arguments that are required to run the script.
        By default, the script has limited access to filesystem and environment,
        but you can request for more access to the environment, filesystem, and syscalls
        by setting the system_access flag to True.

        Note
        -----
        The user must have rootkit access to request system access.

        Parameters
        ----------
        script : str | Path
            The draft script for the custom facet.
            Can either be the text of the script or
            a path to a file containing the script.

        requirements : list[str] | Path | None
            The requirements for the custom facet.
            Can either be a path to a file containing the requirements in pip format,
            or be a list of strings each representing a single requirement.

        system_access : bool, optional
            Whether to grant system access to the custom facet, by default False.
            If True, the custom facet will have unrestricted access to the
            python environment and libraries.
            Creating a custom facet with system access will fail if
            the user does not have rootkit access.

        arguments : dict[str, str | int | float | bool]
            The arguments for the custom facet script and their default values.

        Returns
        -------
        Self
            The updated CustomFacet instance.
        """
        if isinstance(script, Path):
            with script.open("r") as fp:
                script = fp.read()
        script = textwrap.dedent(script)
        # Validate the script syntax againt ikigai platform's python version
        # IPLT-11330: See if we can avoid hardcoding the python version
        ast.parse(script, feature_version=(3, 10))

        if requirements is None:
            requirements = []
        if isinstance(requirements, Path):
            with requirements.open("r") as fp:
                requirements = fp.readlines()

        if arguments is None:
            arguments = {}
        new_arguments = {
            name: CustomFacetArgumentSpec(
                name=name,
                argument_type=CustomFacetArgumentType.from_value(value),
                value=value,
            )
            for name, value in arguments.items()
        }

        # If rootkit is required, generate it from the script
        rootkit_token = (
            self.__client.access.generate_rootkit_token(script=script)
            if system_access
            else ""
        )

        # Update the custom facet
        arguments_list = [argument.to_dict() for argument in new_arguments.values()]

        self.__client.component.edit_custom_facet(
            custom_facet_id=self.custom_facet_id,
            chain_group=self.facet_type.facet_info.chain_group,
            python_script=script,
            libraries=requirements,
            rootkit_token=rootkit_token,
            arguments=arguments_list,
        )

        self.script = script
        self.requirements = requirements
        self.rootkit_token = rootkit_token
        self.arguments = new_arguments
        return self

    @cached_property
    def access(self) -> CustomFacetAccess:
        """
        Manage access to the Custom Facet.

        Returns
        -------
        CustomFacetAccess
            Access manager for the Custom Facet.
        """
        return CustomFacetAccess(
            custom_facet_id=self.custom_facet_id, client=self.__client
        )

    def describe(self) -> datax.CustomFacetDict:
        return self.__client.component.get_custom_facet(
            custom_facet_id=self.custom_facet_id
        )

    def delete(self) -> None:
        self.__client.component.delete_custom_facet(
            custom_facet_id=self.custom_facet_id
        )
        return None

    def create_version(self, name: str) -> CustomFacetVersion:
        arguments_list = [argument.to_dict() for argument in self.arguments.values()]

        custom_facet_version_id = self.__client.component.create_custom_facet_version(
            custom_facet_id=self.custom_facet_id,
            version=name,
            description=self.description,
            python_script=self.script,
            libraries=self.requirements,
            rootkit_token=self.rootkit_token,
            arguments=arguments_list,
        )

        custom_facet_version_dict = self.__client.component.get_custom_facet_version(
            custom_facet_id=self.custom_facet_id,
            version_id=custom_facet_version_id,
        )

        return CustomFacetVersion.from_dict(
            data=custom_facet_version_dict,
            facet_type=self.facet_type,
            client=self.__client,
        )

    def versions(self) -> NamedMapping[CustomFacetVersion]:
        version_dicts = self.__client.component.get_custom_facet_versions(
            custom_facet_id=self.custom_facet_id
        )
        versions = {
            version.version_id: version
            for version in (
                CustomFacetVersion.from_dict(
                    data=version_dict, facet_type=self.facet_type, client=self.__client
                )
                for version_dict in version_dicts
            )
        }
        return NamedMapping(versions)

    def unpinned(self) -> CustomFacetVersion:
        latest_version = next(
            iter(
                sorted(
                    self.versions().values(), key=lambda x: x.created_at, reverse=True
                )
            ),
            None,
        )

        description = latest_version.description if latest_version else self.description
        arguments = [
            argument.to_dict()
            for argument in (
                latest_version.arguments if latest_version else self.arguments
            ).values()
        ]
        created_at = str(
            int(
                (
                    latest_version.created_at if latest_version else self.created_at
                ).timestamp()
            )
        )

        return CustomFacetVersion.from_dict(
            data={
                "version": "",
                "version_id": "",
                "custom_facet_id": self.custom_facet_id,
                "description": description,
                "arguments": arguments,
                "created_at": created_at,
            },
            facet_type=self.facet_type,
            client=self.__client,
        )


class CustomFacetBrowser(ComponentBrowser[CustomFacet]):
    __client: Client

    def __init__(self, client: Client) -> None:
        self.__client = client

    @deprecated(
        "Prefer directly loading by name:\n\tikigai.custom_facets['custom_facet_name']"
    )
    def __call__(self) -> NamedMapping[CustomFacet]:
        custom_facets = {
            custom_facet["custom_facet_id"]: CustomFacet.from_dict(
                data=custom_facet, client=self.__client
            )
            for custom_facet in self.__client.component.get_custom_facets_for_user()
        }

        return NamedMapping(custom_facets)

    def __getitem__(self, name: str) -> CustomFacet:
        custom_facet_dict = self.__client.component.get_custom_facet_by_name(name=name)

        return CustomFacet.from_dict(data=custom_facet_dict, client=self.__client)

    def search(self, query: str) -> NamedMapping[CustomFacet]:
        matching_custom_facets = {
            custom_facet["custom_facet_id"]: CustomFacet.from_dict(
                data=custom_facet, client=self.__client
            )
            for custom_facet in self.__client.search.search_custom_facets_for_user(
                query=query
            )
        }

        return NamedMapping(matching_custom_facets)


class CustomFacetVersion(BaseModel):
    name: str = Field(validation_alias=AliasChoices("name", "version"))
    version_id: str
    custom_facet_id: str
    description: str
    arguments: dict[str, CustomFacetArgumentSpec]
    created_at: datetime
    __facet_type: CustomFacetType = PrivateAttr()
    __client: Client = PrivateAttr()

    @field_validator("arguments", mode="before")
    @classmethod
    def validate_arguments(cls, v: list[dict]) -> dict[str, CustomFacetArgumentSpec]:
        if not isinstance(v, list):
            error_msg = "Expected a list of argument dictionaries"
            raise ValueError(error_msg)

        return {
            argument["name"]: CustomFacetArgumentSpec.model_validate(argument)
            for argument in v
        }

    @classmethod
    def from_dict(
        cls, data: datax.CustomFacetVersionDict, facet_type: FacetType, client: Client
    ) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__facet_type = CustomFacetType.from_facet_type(
            facet_type=facet_type,
            custom_facet_id=self.custom_facet_id,
            version_id=self.version_id,
            custom_facet_argument_specs=self.arguments,
        )
        self.__client = client
        return self

    @property
    def facet_type(self) -> CustomFacetType:
        return self.__facet_type



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from dataclasses import InitVar
from http import HTTPStatus
from typing import Annotated, Any, TypeAlias

import requests
from pydantic import AnyUrl, ConfigDict, EmailStr, Field
from pydantic.dataclasses import dataclass
from requests import Response

from ikigai.utils.compatibility import HTTPMethod

logger = logging.getLogger("ikigai.client")


PEMfilePath: TypeAlias = Annotated[
    str, "Path to a PEM file containing SSL certificates"
]
CertKeyPair: TypeAlias = Annotated[
    tuple[str, str], "Tuple containing certificate and key"
]
SSLConfig: TypeAlias = bool | PEMfilePath | CertKeyPair


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class Session:
    # Init only vars
    user_email: InitVar[EmailStr]
    api_key: InitVar[str]
    ssl: InitVar[SSLConfig]

    base_url: AnyUrl
    __session: requests.Session = Field(init=False)

    def __post_init__(self, user_email: EmailStr, api_key: str, ssl: SSLConfig) -> None:
        self.__session = requests.Session()
        if isinstance(ssl, bool):
            self.__session.verify = ssl
        else:
            self.__session.cert = ssl
        self.__session.headers.update({"user": user_email, "api-key": api_key})

    def request(
        self,
        method: HTTPMethod,
        path: str,
        params: dict[str, str] | None = None,
        json: dict | None = None,
        *,
        suppress_logging: bool = False,
    ) -> Response:
        logger.debug(
            "[%(method)s] %(path)s %(params)s\njson: %(json)s",
            {"method": method, "path": path, "params": params, "json": json},
        )
        url = f"{self.base_url}{path}"
        resp = self.__session.request(
            method=method,
            url=url,
            params=params,
            json=json,
        )
        if resp.status_code < HTTPStatus.BAD_REQUEST:
            return resp
        if resp.status_code < HTTPStatus.INTERNAL_SERVER_ERROR:
            # A 4XX error happened
            if not suppress_logging:
                logger.error(
                    "Request"
                    "[%(method)s] %(path)s %(params)s\n"
                    "%(request)s\n\n"
                    "Response [%(status)s]"
                    "headers: %(response_headers)s\n"
                    "%(response)s\n\n",
                    {
                        "method": method,
                        "path": path,
                        "params": params,
                        "request": resp.request.body,
                        "status": resp.status_code,
                        "response_headers": resp.headers,
                        "response": resp.text,
                    },
                )
            message = (
                f"[{resp.status_code}] Server rejected the request. "
                "Check the request and try again.\n"
                f"Response: {resp.text}"
            )
            raise RuntimeError(message)

        # A 5XX error happened
        if not suppress_logging:
            logger.error(
                "Request"
                "[%(method)s] %(path)s %(params)s\n"
                "%(request)s\n\n"
                "Response [%(status)s]"
                "headers: %(response_headers)s\n"
                "%(response)s\n\n",
                {
                    "method": method,
                    "path": path,
                    "params": params,
                    "request": resp.request.body,
                    "status": resp.status_code,
                    "response_headers": resp.headers,
                    "response": resp.text,
                },
            )
        message = (
            f"[{resp.status_code}] The server encountered an error. "
            "Try again later, if problem persists report an issue.\n"
            f"Response: {resp.text}"
        )
        raise RuntimeError(message)

    def get(
        self,
        path: str,
        params: dict[str, Any] | None = None,
        *,
        suppress_logging: bool = False,
    ) -> Response:
        return self.request(
            method=HTTPMethod.GET,
            path=path,
            params=params,
            suppress_logging=suppress_logging,
        )

    def post(
        self,
        path: str,
        json: dict[Any, Any] | None = None,
        *,
        suppress_logging: bool = False,
    ) -> Response:
        return self.request(
            method=HTTPMethod.POST,
            path=path,
            json=json,
            suppress_logging=suppress_logging,
        )

    def __del__(self) -> None:
        self.__session.close()



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import time
from collections.abc import Mapping
from datetime import datetime
from typing import Any, TypeVar, cast

from pydantic import (
    AliasChoices,
    BaseModel,
    EmailStr,
    Field,
    PrivateAttr,
    field_validator,
)
from tqdm.auto import tqdm

from ikigai.client import Client, datax
from ikigai.components._flow_definition_shim import flow_versioning_shim
from ikigai.components.flow_definition import FlowDefinition
from ikigai.typing import ComponentBrowser, Directory, NamedDirectoryDict, NamedMapping
from ikigai.typing.pydantic_extensions import (
    CronStr,
    OptionalStr,
    TimestampSerializableDatetime,
    TimestampSerializableOptionalDatetime,
)
from ikigai.utils import DirectoryType, FlowStatus
from ikigai.utils.compatibility import Self, deprecated, override

logger = logging.getLogger("ikigai.components")

T = TypeVar("T")


class Schedule(BaseModel):
    """
    Schedule for a flow.
    """

    name: str
    """Name of the schedule."""
    start_time: TimestampSerializableDatetime
    """Start time of the schedule."""
    end_time: TimestampSerializableOptionalDatetime = None
    """End time of the schedule. If None, the schedule will run indefinitely."""
    cron: CronStr
    """Cron expression for the schedule."""

    @field_validator("end_time", "start_time", mode="before")
    @classmethod
    def validate_time(cls, v: object) -> datetime | str | None:
        if v is None or v == "":
            return None

        if not isinstance(v, (str, datetime)):
            error_msg = f"Must be a timestamp, datetime, or None. Got {type(v)}"
            raise TypeError(error_msg)

        if isinstance(v, datetime):
            # Remove microseconds resolution
            v = v.replace(microsecond=0)
            # If the datetime is naive, convert it to local timezone
            if v.tzinfo is None:
                v = v.astimezone()

        return v

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        return cls.model_validate(data)

    def to_dict(self) -> datax.ScheduleDict:
        return cast(datax.ScheduleDict, self.model_dump())


class FlowBuilder:
    _app_id: str
    _name: str
    _directory: Directory | None
    _high_volume_preference: bool
    _flow_definition: datax.FlowDefinitionDict
    _schedule: Schedule | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        """
        Initialize the FlowBuilder

        Note
        ----
        The FlowBuilder initialization should only happen internally via
        the App object.

        Parameters
        ----------
        client : Client
            The Ikigai client to use for API interactions

        app_id : str
            The ID of the App to which the flow will belong

        Returns
        -------
        None
        """
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._directory = None
        self._high_volume_preference = False
        self._flow_definition = FlowDefinition().to_dict()
        self._schedule = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def definition(
        self, definition: Flow | FlowDefinition | datax.FlowDefinitionDict
    ) -> Self:
        if isinstance(definition, FlowDefinition):
            self._flow_definition = definition.to_dict()
            return self

        if isinstance(definition, Flow):
            if definition.app_id != self._app_id:
                error_msg = (
                    "Building flow from a diferent app is not supported\n"
                    "source_app != destination_app "
                    f"({definition.app_id} != {self._app_id})"
                )
                raise ValueError(error_msg)
            flow_dict = self.__client.component.get_flow(flow_id=definition.flow_id)
            self._flow_definition = flow_dict["definition"]
            return self

        if isinstance(definition, dict):
            self._flow_definition = definition
            return self

        error_msg = (
            f"Definition was of type {type(definition)} but, "
            "must be a Flow or FlowDefinition or FlowDefinitionDict"
        )
        raise TypeError(error_msg)

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def high_volume_preference(self, optimize: bool) -> Self:
        """
        Set the high volume preference flag for the flow.

        Parameters
        ----------
        optimize : bool
            The high volume preference to set for the flow.
            True if the flow should be optimized for high volume data processing,
            False otherwise.

        Returns
        -------
        Self
            The FlowBuilder instance with high volume preference set.
        """
        self._high_volume_preference = optimize
        return self

    def schedule(
        self,
        schedule: Schedule | datax.ScheduleDict | str,
    ) -> Self:
        """
        Set the schedule for the flow.

        Parameters
        ----------
        schedule : Schedule | ScheduleDict | str
            The schedule to set for the flow. Can be provided as a Schedule object,
            a Schedule dictionary, or a cron string.

        Returns
        -------
        Self
            The FlowBuilder instance with schedule set.
        """
        if isinstance(schedule, str):
            self._schedule = Schedule(
                name=self._name,
                cron=schedule,
                start_time=datetime.now(),
                end_time=None,
            )
            return self

        if isinstance(schedule, Mapping):
            self._schedule = Schedule.from_dict(schedule)
            return self

        if isinstance(schedule, Schedule):
            self._schedule = schedule
            return self

        error_msg = (
            f"Schedule was of type {type(schedule)} but, "
            "must be a Schedule, ScheduleDict, or cron string"
        )
        raise TypeError(error_msg)

    def build(self) -> Flow:
        """
        Build the Flow object

        Creates the flow in the Ikigai platform using the provided
        parameters and returns the corresponding Flow object.

        Returns
        -------
        Flow
            The created Flow object
        """
        flow_id = self.__client.component.create_flow(
            app_id=self._app_id,
            name=self._name,
            directory=self._directory,
            high_volume_preference=self._high_volume_preference,
            flow_definition=self._flow_definition,
            schedule=self._schedule.to_dict() if self._schedule else None,
        )

        # Populate Flow object
        flow_dict = self.__client.component.get_flow(flow_id=flow_id)

        return Flow.from_dict(data=flow_dict, client=self.__client)


class FlowStatusReport(BaseModel):
    status: FlowStatus
    progress: int | None = Field(default=None)
    message: str

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        return cls.model_validate(data)


class RunLog(BaseModel):
    log_id: str
    status: FlowStatus
    user: EmailStr
    erroneous_facet_id: OptionalStr
    data: str = Field(validation_alias="message")
    timestamp: datetime

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        return cls.model_validate(data)


class Flow(BaseModel):
    app_id: str = Field(validation_alias=AliasChoices("app_id", "project_id"))
    flow_id: str = Field(validation_alias=AliasChoices("flow_id", "pipeline_id"))
    name: str
    schedule: Schedule | None = None
    created_at: datetime
    modified_at: datetime
    __client: Client = PrivateAttr()

    @field_validator("schedule", mode="before")
    @classmethod
    def validate_schedule(cls, v: T) -> T | None:
        if isinstance(v, Mapping) and not any(bool(value) for value in v.values()):
            # If the schedule returned is filled with blank values,
            #   it means the schedule is not set.
            #   This is a quirk of grpc serialization.
            return None
        return v

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> dict:
        return {
            "flow_id": self.flow_id,
            "name": self.name,
            "schedule": self.schedule.to_dict() if self.schedule else None,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
        }

    def delete(self) -> None:
        self.__client.component.delete_flow(app_id=self.app_id, flow_id=self.flow_id)
        return None

    def rename(self, name: str) -> Self:
        self.__client.component.edit_flow(
            app_id=self.app_id, flow_id=self.flow_id, name=name
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def update_schedule(
        self, schedule: Schedule | datax.ScheduleDict | str | None = None
    ) -> Self:
        """
        Update the schedule for the flow.

        Parameters
        ----------
        schedule : Schedule | ScheduleDict | str | None
            The schedule to set for the flow.
            Can be provided as a Schedule object, a Schedule dictionary, or a
            cron string. Pass None to remove existing schedule.

        Returns
        -------
        Self
            The updated Flow object.
        """
        if isinstance(schedule, str):
            schedule = Schedule(
                name=self.name, cron=schedule, start_time=datetime.now(), end_time=None
            )
        if isinstance(schedule, Mapping):
            schedule = Schedule.from_dict(schedule)

        self.__client.component.edit_flow(
            app_id=self.app_id,
            flow_id=self.flow_id,
            schedule=schedule.to_dict() if schedule else None,
        )

        if schedule is None:
            # HACK: BE has no way to remove schedule via API,
            #   so we wait for 90 seconds to ensure the schedule is removed
            time.sleep(90)

        # Update the schedule attribute
        self.schedule = schedule
        return self

    def move(self, directory: Directory) -> Self:
        self.__client.component.edit_flow(
            app_id=self.app_id, flow_id=self.flow_id, directory=directory
        )
        return self

    def update_high_volume_preference(self, optimize: bool) -> Self:
        """
        Update the high volume preference of the flow.

        Parameters
        ----------
        optimize : bool
            The new high volume preference to set for the flow.
            True if the flow should be optimized for high volume data processing,
            False otherwise.

        Returns
        -------
        Self
            The updated Flow object.
        """
        self.__client.component.edit_flow(
            app_id=self.app_id,
            flow_id=self.flow_id,
            high_volume_preference=optimize,
        )
        return self

    def update_definition(
        self, definition: FlowDefinition | datax.FlowDefinitionDict
    ) -> Self:
        """
        Update the flow definition.

        Replaces the existing flow definition with the provided one.

        Parameters
        ----------
        definition : FlowDefinition | FlowDefinitionDict
            The new flow definition to set. Can be provided as a FlowDefinition
            object or as a dictionary.

        Returns
        -------
        Self
            The updated Flow object.

        """
        if isinstance(definition, FlowDefinition):
            definition = definition.to_dict()

        self.__client.component.edit_flow(
            app_id=self.app_id, flow_id=self.flow_id, flow_definition=definition
        )
        return self

    def status(self) -> FlowStatusReport:
        resp = self.__client.component.is_flow_runing(
            app_id=self.app_id, flow_id=self.flow_id
        )
        return FlowStatusReport.from_dict(resp)

    def run_logs(
        self, max_count: int = 1, since: datetime | None = None
    ) -> list[RunLog]:
        log_dicts = self.__client.component.get_flow_log(
            app_id=self.app_id, flow_id=self.flow_id, max_count=max_count
        )

        run_logs = [RunLog.from_dict(data=log) for log in log_dicts]
        if since is not None:
            run_logs = [log for log in run_logs if log.timestamp > since]
        return run_logs

    def run(self, **variables) -> RunLog:
        """
        Run the flow

        This function will run the flow and wait for it to complete.
        It will accept any number of keyword arguments which will be passed as run
        variables to the flow.

        Parameters
        ----------
        **variables : dict
            Run variables to be passed to the flow, where the key is the variable name
            and the value is the variable value. Keys that start with an
            underscore ("_") are reserved for future use and will be ignored.

        Returns
        -------
        RunLog
            The final run log of the flow after completion
        """
        run_variables: datax.RunVariablesRequest = {
            key: {"value": value}
            for key, value in variables.items()
            if not key.startswith("_")
        }

        # Start running pipeline
        self.__client.component.run_flow(
            app_id=self.app_id, flow_id=self.flow_id, variables=run_variables
        )

        return self.__await_run()

    def describe(self) -> datax.FlowDict:
        flow = self.__client.component.get_flow(flow_id=self.flow_id)
        # Apply flow_versioning_shim to allow migration of older flows
        # TODO: Remove this shim after "important" flows are migrated
        facet_specs = self.__client.get(
            path="/component/get-facet-specs",
        ).json()
        # shim is a hack so better to keep it explicit
        shimed_flow = flow_versioning_shim(flow=flow, facet_specs=facet_specs)
        return shimed_flow  # noqa: RET504

    def __await_run(self) -> RunLog:
        start_time = datetime.now().astimezone()
        # TODO: Switch to using websockets once they are available
        with tqdm(total=100, dynamic_ncols=True) as progress_bar:
            status_report = self.status()
            progress_bar.desc = status_report.status
            progress_bar.update(0)

            # Initially wait while pipeline is scheduled
            while status_report.status == FlowStatus.SCHEDULED:
                time.sleep(5)
                status_report = self.status()

            last_progress = status_report.progress if status_report.progress else 0
            progress_bar.desc = status_report.status
            progress_bar.update(last_progress)

            # Wait while pipeline is running
            running_states = (
                FlowStatus.RUNNING,  # Flow is running
                FlowStatus.STOPPING,  # Flow is in the process of stopping
                FlowStatus.SCHEDULED,  # Flow is scheduled (again), likely a retry
                FlowStatus.UNKNOWN,  # Flow status is unknown, but known was running
            )
            while status_report.status in running_states:
                time.sleep(1)
                status_report = self.status()
                progress = status_report.progress if status_report.progress else 100
                progress_bar.desc = status_report.status
                new_progress = last_progress + max(progress - last_progress, 0)
                progress_bar.update(new_progress - last_progress)
                last_progress = new_progress
            # Flow run completed

            # Get status from logs and update progress bar
            run_logs = self.run_logs(max_count=1, since=start_time)
            if not run_logs:
                # TODO: Give a better error message
                error_msg = (
                    "No logs found for"
                    f" <Flow(flow_id={self.flow_id}, name={self.name})>"
                    f" after the flow started running ({start_time=})."
                )
                raise RuntimeError(error_msg)
            run_log = run_logs[0]

            progress = 100
            progress_bar.desc = run_log.status
            progress_bar.update(progress - last_progress)

            return run_log


class FlowBrowser(ComponentBrowser[Flow]):
    __app_id: str
    __client: Client

    def __init__(self, app_id: str, client: Client) -> None:
        self.__app_id = app_id
        self.__client = client

    @deprecated("Prefer directly loading by name:\n\tapp.flows['flow_name']")
    @override
    def __call__(self) -> NamedMapping[Flow]:
        flows = {
            flow["pipeline_id"]: Flow.from_dict(data=flow, client=self.__client)
            for flow in self.__client.component.get_flows_for_app(app_id=self.__app_id)
        }

        return NamedMapping(flows)

    @override
    def __getitem__(self, name: str) -> Flow:
        flow_dict = self.__client.component.get_flow_by_name(
            app_id=self.__app_id, name=name
        )

        return Flow.from_dict(data=flow_dict, client=self.__client)

    @override
    def search(self, query: str) -> NamedMapping[Flow]:
        matched_flows = {
            flow["pipeline_id"]: Flow.from_dict(data=flow, client=self.__client)
            for flow in self.__client.search.search_flows_for_project(
                app_id=self.__app_id, query=query
            )
        }

        return NamedMapping(matched_flows)


class FlowDirectoryBuilder:
    _app_id: str
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._parent = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def parent(self, parent: Directory) -> Self:
        self._parent = parent
        return self

    def build(self) -> FlowDirectory:
        directory_id = self.__client.component.create_flow_directory(
            app_id=self._app_id, name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_flow_directory(
            app_id=self._app_id, directory_id=directory_id
        )

        return FlowDirectory.from_dict(data=directory_dict, client=self.__client)


class FlowDirectory(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    directory_id: str
    name: str
    __client: Client = PrivateAttr()

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.FLOW

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_flow_directories_for_app(
            app_id=self.app_id, parent=self
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def flows(self) -> NamedMapping[Flow]:
        flow_dicts = self.__client.component.get_flows_for_app(
            app_id=self.app_id, directory_id=self.directory_id
        )

        flows = {
            flow.flow_id: flow
            for flow in (
                Flow.from_dict(data=flow_dict, client=self.__client)
                for flow_dict in flow_dicts
            )
        }

        return NamedMapping(flows)



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ikigai.client import datax


def flow_versioning_shim(flow: datax.FlowDict, facet_specs: dict) -> datax.FlowDict:
    """
    Shim to improve compatibility with older Flow Definitions
    and migrate them to the latest format.

    Parameters
    ----------
    flow: FlowDict
        The flow dict to shim for improved compatibility with
        latest flow definition format

    facet_specs: dict
        The facet specifications dict to use as reference for flow definition format

    Returns
    -------
    FlowDict
        The shimed flow dict
    """
    facets = flow["definition"].get("facets", [])
    flow["definition"]["facets"] = [
        _flow_facet_shim(
            facet=facet, facet_spec=__search_facet_spec(facet_specs, facet["facet_uid"])
        )
        for facet in facets
    ]

    return flow


def _flow_facet_shim(
    facet: datax.FacetDict, facet_spec: dict | None
) -> datax.FacetDict:
    """
    Shim to improve compatibility with older Flow Definitions
    Fix the facet structure to match the expected format

    Parameters
    ----------
    facet: FacetDict
        The facet dict to shim for improved compatibility

    Returns
    -------
    FacetDict
        The shimed facet dict
    """
    # Initial cleaning of facet args
    facet_args = facet.get("arguments", {})
    if "submodel_type" in facet_args:
        facet_args["sub_model_type"] = facet_args["submodel_type"]
        del facet_args["submodel_type"]

    facet["arguments"] = facet_args.copy()

    if facet_spec is None:
        # If no facet specification is found, return the facet as is
        return facet

    # Normalize the facet arguments based on the facet specification
    for key, value in facet["arguments"].items():
        argument_spec = next(
            (
                arg_spec
                for arg_spec in facet_spec["facet_arguments"]
                if arg_spec["name"] == key
            ),
            None,
        )
        if not argument_spec:
            # If the argument is not specified in the facet spec, skip it
            del facet_args[key]
            continue

        normalization_is_required: bool = (
            isinstance(value, list)
            and not argument_spec.get("is_list", False)
            and argument_spec["argument_type"] == "MAP"
        )
        if normalization_is_required:
            facet_args[key] = value[0] if value else {}

        if "experiment_selection" in facet_args:
            del facet_args["experiment_selection"]

    # Update the facet arguments with the normalized values
    facet["arguments"] = facet_args
    return facet


def __search_facet_spec(facet_specs: dict, facet_uid: str) -> dict | None:
    """
    Search for a facet specification by facet UID

    Parameters
    ----------
    facet_specs: dict
        The facet specifications dict to search in

    facet_uid: str
        The UID of the facet to search for

    Returns
    -------
        dict or None
            The facet specification dict if found, otherwise None
    """
    for chain_group in facet_specs.values():
        for facet_group in chain_group.values():
            for facet_type in facet_group.values():
                if facet_type["facet_info"]["facet_uid"] == facet_uid:
                    return facet_type

    # Facet specification with facet_uid not found in facet specifications
    return None



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from contextlib import ExitStack

import pytest

from ikigai import Ikigai


def test_app_creation(ikigai: Ikigai, app_name: str, cleanup: ExitStack) -> None:
    apps = ikigai.apps()
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    app_dict = app.to_dict()
    apps_after_creation = ikigai.apps()

    with pytest.raises(KeyError):
        apps.get_id(app.app_id)
    assert app_dict["name"] == app_name
    assert app_dict["description"] == "A test app"
    assert apps_after_creation.get_id(app.app_id) is not None


def test_app_editing(ikigai: Ikigai, app_name: str, cleanup: ExitStack) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    app.rename(f"updated {app_name}")
    app.update_description("An updated test app")

    app_after_edit = ikigai.apps().get_id(id=app.app_id)

    assert app_after_edit.name == app.name
    assert app_after_edit.description == app.description
    assert app_after_edit.name == f"updated {app_name}"
    assert app_after_edit.description == "An updated test app"


def test_app_describe_1(ikigai: Ikigai, app_name: str, cleanup: ExitStack) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    description = app.describe()
    assert description is not None

    assert "app" in description
    assert description["app"]["name"] == app_name

    assert "components" in description
    components = description["components"]

    assert "charts" in components
    assert components["charts"] == []

    assert "connectors" in components
    assert components["connectors"] == []

    assert "dashboards" in components
    assert components["dashboards"] == []

    assert "datasets" in components
    assert components["datasets"] == []

    assert "databases" in components
    assert components["databases"] == []

    assert "pipelines" in components
    assert components["pipelines"] == []

    assert "models" in components
    assert components["models"] == []

    assert "external_resources" in components
    assert components["external_resources"] == []


def test_app_directory(
    ikigai: Ikigai, app_directory_name: str, cleanup: ExitStack
) -> None:
    app_directory = ikigai.app_directory.new(name=app_directory_name).build()
    cleanup.callback(app_directory.delete)

    app_directories = ikigai.directories()

    assert len(app_directories) >= 1
    assert app_directory_name in app_directories
    fetched_app_directory = app_directories[app_directory_name]
    assert fetched_app_directory.directory_id == app_directory.directory_id


def test_app_browser_1(ikigai: Ikigai, app_name: str, cleanup: ExitStack) -> None:
    app = ikigai.app.new(name=app_name).description("Get by name").build()
    cleanup.callback(app.delete)

    fetched_app = ikigai.apps[app_name]
    assert fetched_app.app_id == app.app_id
    assert fetched_app.name == app_name


def test_app_browser_search_1(
    ikigai: Ikigai, app_name: str, cleanup: ExitStack
) -> None:
    app = ikigai.app.new(name=app_name).description("Get by name").build()
    cleanup.callback(app.delete)

    app_name_substr = app_name.split("-", maxsplit=1)[1]
    fetched_apps = ikigai.apps.search(app_name_substr)

    assert app_name in fetched_apps
    fetched_app = fetched_apps[app_name]

    assert fetched_app.app_id == app.app_id
    assert fetched_app.name == app.name == app_name



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

import logging
import random
import sys
from collections.abc import Generator
from contextlib import ExitStack
from pathlib import Path
from typing import Any, cast

import pytest
from _pytest.fixtures import FixtureRequest

# Multiple python version compatible import for reading toml
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def _read_credentials(env_file: Path) -> dict[str, Any]:
    with env_file.open("rb") as env:
        users: dict[str, dict[str, str]] = tomllib.load(env)["credentials"]["users"]
    return {"ids": users.keys(), "params": users.values()}


@pytest.fixture(**_read_credentials(Path("./test-env.toml")))
def cred(request: FixtureRequest) -> dict[str, str]:
    assert isinstance(request.param, dict)
    return cast(dict[str, str], request.param)


@pytest.fixture()
def logger() -> logging.Logger:
    return logging.getLogger("test")


_ALPHABET = "abcdefghijklmnopqrstuvwxyz0123456789-"


@pytest.fixture()
def random_name() -> str:
    name_length = int(random.triangular(low=5, high=20, mode=20))
    return "".join(random.choices(_ALPHABET, k=name_length))


@pytest.fixture()
def random_name_2() -> str:
    name_length = int(random.triangular(low=5, high=20, mode=20))
    return "".join(random.choices(_ALPHABET, k=name_length))


@pytest.fixture()
def random_name_3() -> str:
    name_length = int(random.triangular(low=5, high=20, mode=20))
    return "".join(random.choices(_ALPHABET, k=name_length))


@pytest.fixture()
def cleanup() -> Generator[ExitStack, None, None]:
    exit_stack = ExitStack()
    try:
        yield exit_stack
    finally:
        exit_stack.close()



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from contextlib import ExitStack

import pandas as pd
import pytest

from ikigai import FlowStatus, Ikigai


def test_flow_definition_builder_facet_types(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    assert facet_types.INPUT
    assert facet_types.MID
    assert facet_types.OUTPUT

    # Assorted facet types tests
    assert "PYTHON" in facet_types.INPUT
    assert "PYTHON" in facet_types.MID
    assert "PYTHON" in facet_types.OUTPUT

    assert "IMPORTED" in facet_types.INPUT
    assert "IMPORTED" not in facet_types.MID
    assert "IMPORTED" not in facet_types.OUTPUT

    assert "EXPORTED" not in facet_types.INPUT
    assert "EXPORTED" not in facet_types.MID
    assert "EXPORTED" in facet_types.OUTPUT


def test_flow_definition_empty(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flows = app.flows()
    assert len(flows) == 0

    flow = app.flow.new(name=flow_name).definition(ikigai.builder.build()).build()

    assert flow.flow_id
    assert flow.name == flow_name


def test_flow_definition_simple(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    r"""
    Flow:
    Import -> Count -> Export
    """
    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.MID.COUNT)
        .arguments(
            output_column_name="count",
            sort=True,
            target_columns=df1.columns.tolist()[:-2],
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED)
        .arguments(dataset_name=f"output-{flow_name}", file_type="csv", header=True)
        .build()
    )
    flow = app.flow.new(name=flow_name).definition(flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data


def test_flow_definition_multiple_parent_facets(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    dataset1 = app.dataset.new(name=f"{dataset_name}-1").df(df1).build()

    dataset2 = app.dataset.new(name=f"{dataset_name}-2").df(df2).build()

    r"""
    Flow:
    Import-1 -\
               >-> Union -> Export
    Import-2 -/
    """
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    imported_facet_1 = builder.facet(
        facet_type=facet_types.INPUT.IMPORTED, name="imported-facet-1"
    ).arguments(
        dataset_id=dataset1.dataset_id,
        file_type="csv",
        header=True,
        use_raw_file=False,
    )
    imported_facet_2 = builder.facet(
        facet_type=facet_types.INPUT.IMPORTED, name="imported-facet-2"
    ).arguments(
        dataset_id=dataset2.dataset_id,
        file_type="csv",
        header=True,
        use_raw_file=False,
    )
    flow_definition = (
        builder.facet(facet_type=facet_types.MID.UNION, name="union")
        .add_arrow(imported_facet_1, table_side="top")
        .add_arrow(imported_facet_2, table_side="bottom")
        .arguments(option="full")
        .facet(facet_type=facet_types.OUTPUT.EXPORTED)
        .arguments(dataset_name=f"output-{flow_name}", file_type="csv", header=True)
        .build()
    )
    flow = app.flow.new(name=flow_name).definition(flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data


def test_flow_definition_multiple_child_facets(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    r"""
    Flow:
                     /-> Count -> Export
    Import -> Copy -<
                     \-> Export
    """
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    source_data = (
        builder.facet(facet_type=facet_types.INPUT.IMPORTED, name="imported-facet")
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.MID.COPY)
    )
    (
        source_data.facet(facet_type=facet_types.MID.COUNT)
        .arguments(target_columns=df1.columns.tolist()[:-2])
        .facet(facet_type=facet_types.OUTPUT.EXPORTED)
        .arguments(
            dataset_name=f"output-{flow_name}-count", file_type="csv", header=True
        )
    )
    source_data.facet(facet_type=facet_types.OUTPUT.EXPORTED).arguments(
        dataset_name=f"output-{flow_name}", file_type="csv", header=True
    )

    flow_definition = builder.build()
    flow = app.flow.new(name=flow_name).definition(flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data


def test_flow_definition_simple_ml_facet(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df_ml_regression1: pd.DataFrame,
    model_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df_ml_regression1).build()

    model_types = ikigai.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types["Linear"]["Lasso"],
        )
        .arguments(
            model_name=model.name,
            version="initial-train",
        )
        .hyperparameters(alpha=0.1)
        .parameters(
            target_column="target",
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED)
        .arguments(dataset_name=f"output-{flow_name}", file_type="csv", header=True)
        .build()
    )
    flow = app.flow.new(name=flow_name).definition(flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data


def test_flow_definition_facet_argument_typechecking_scalar_1(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    with pytest.raises(TypeError):
        # dataset_id should be str, not int
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).arguments(
            dataset_id=123,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )


def test_flow_definition_facet_argument_typechecking_scalar_2(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    with pytest.raises(TypeError):
        # header should be bool, not str
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).arguments(
            dataset_id="123",
            file_type="csv",
            header="True",
            use_raw_file=False,
        )


def test_flow_definition_variables_check_1(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="already exists for another facet"):
        builder.facet(facet_type=facet_types.INPUT.IMPORTED, name="input").variables(
            dataset="dataset_id",
        ).facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output").variables(
            dataset="dataset_name",
        )


def test_flow_definition_variables_check_2(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="Please set a name for the facet"):
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).variables(
            dataset="dataset_id",
        )


def test_flow_definition_variables_check_3(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="does not have argument 'bad_arg_name'"):
        builder.facet(facet_type=facet_types.INPUT.IMPORTED, name="input").variables(
            dataset="bad_arg_name",
        )



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from contextlib import ExitStack
from logging import Logger

import pandas as pd
import pytest

from ikigai import FlowStatus, Ikigai
from ikigai.components.dataset import _get_dataset_download_url
from ikigai.utils import DatasetDownloadStatus


def test_dataset_creation(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    datasets = app.datasets()
    assert len(datasets) == 0

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    with pytest.raises(KeyError):
        datasets.get_id(dataset.dataset_id)

    datasets_after_creation = app.datasets()
    assert len(datasets_after_creation) == 1

    dataset_dict = dataset.to_dict()
    assert dataset_dict["name"] == dataset_name


def test_dataset_editing(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    dataset.rename(f"updated {dataset_name}")
    dataset.edit_data(df2)

    dataset_after_edit = app.datasets().get_id(dataset.dataset_id)
    round_trip_df2 = dataset_after_edit.df()

    assert dataset_after_edit.name == dataset.name
    assert dataset_after_edit.name == f"updated {dataset_name}"
    assert df2.columns.equals(round_trip_df2.columns)
    pd.testing.assert_frame_equal(
        df2, round_trip_df2, check_dtype=False, check_exact=False
    )


def test_dataset_download(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
    logger: Logger,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    round_trip_df1 = dataset.df()
    assert len(df1) == len(round_trip_df1)
    assert df1.columns.equals(round_trip_df1.columns)

    # v. helpful debug message when the test fails
    logger.info(
        ("df1.dtypes:\n%r\n%r\n\nround_trip_df1.dtypes:\n%r\n%r\n\n"),
        df1.dtypes,
        df1.head(),
        round_trip_df1.dtypes,
        round_trip_df1.head(),
    )

    pd.testing.assert_frame_equal(
        df1, round_trip_df1, check_dtype=False, check_exact=False
    )


def test_dataset_async_download(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
    logger: Logger,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    Test async dataset download by running a high volume flow.

    This test creates a high volume pipeline with a simple imported->exported
    structure. The output dataset will require async download processing,
    which tests the DatasetDownloadStatus.IN_PROGRESS handling.
    """
    app = (
        ikigai.app.new(name=app_name).description("Test async dataset download").build()
    )
    cleanup.callback(app.delete)

    # Create a small input dataset
    input_dataset = app.dataset.new(name=dataset_name).df(df1).build()

    # Create a high volume flow with imported->exported structure
    facet_types = ikigai.facet_types
    output_dataset_name = f"output-{dataset_name}"
    flow_definition = (
        ikigai.builder.facet(
            facet_type=facet_types.INPUT.IMPORTED, name=input_dataset.name
        )
        .arguments(
            dataset_id=input_dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(
            dataset_name=output_dataset_name,
            file_type="csv",
            header=True,
        )
        .build()
    )

    # Build flow with high volume preference enabled
    flow = (
        app.flow.new(name=f"flow-{dataset_name}")
        .definition(flow_definition)
        .high_volume_preference(optimize=True)
        .build()
    )

    # Run the flow - this should create an output dataset
    logger.info("Running high volume flow to generate output dataset")
    run_log = flow.run()

    # Verify the flow succeeded
    assert run_log.status == FlowStatus.SUCCESS, run_log.data

    # Get the output dataset
    output_dataset = app.datasets[output_dataset_name]

    # Track whether IN_PROGRESS status was encountered
    in_progress_encountered = False

    # Wrap the original function to track IN_PROGRESS status
    original_get_url = _get_dataset_download_url

    def tracked_get_dataset_download_url(client, app_id, dataset_id):
        nonlocal in_progress_encountered
        # Call initialize_dataset_download to check initial status
        response = client.component.initialize_dataset_download(
            app_id=app_id,
            dataset_id=dataset_id,
        )

        # Track if we got IN_PROGRESS
        if response["status"] == DatasetDownloadStatus.IN_PROGRESS:
            in_progress_encountered = True
            logger.info("✓ Async download triggered: IN_PROGRESS status detected")

        # Delegate to original implementation
        return original_get_url(client, app_id, dataset_id)

    monkeypatch.setattr(
        "ikigai.components.dataset._get_dataset_download_url",
        tracked_get_dataset_download_url,
    )

    # Download the dataset - this will test the async download logic
    logger.info("Attempting to download output dataset (should trigger async download)")
    downloaded_df = output_dataset.df()

    # Verify that IN_PROGRESS status was encountered
    assert in_progress_encountered, (
        "Expected IN_PROGRESS status to be returned during async download, "
        "but it was not encountered"
    )

    # Verify the downloaded data matches the input
    assert len(downloaded_df) == len(df1)
    assert downloaded_df.columns.equals(df1.columns)

    logger.info(
        ("Input df1:\n%r\n%r\n\nDownloaded df:\n%r\n%r\n\n"),
        df1.dtypes,
        df1.head(),
        downloaded_df.dtypes,
        downloaded_df.head(),
    )

    pd.testing.assert_frame_equal(
        df1, downloaded_df, check_dtype=False, check_exact=False
    )


def test_dataset_describe(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)
    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    description = dataset.describe()
    assert description is not None
    assert description["name"] == dataset_name
    assert description["project_id"] == app.app_id
    assert description["directory"] is not None
    assert description["directory"]["type"] == "DATASET"


def test_dataset_directories_creation(
    ikigai: Ikigai,
    app_name: str,
    dataset_directory_name_1: str,
    dataset_directory_name_2: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test dataset directory creation")
        .build()
    )
    cleanup.callback(app.delete)

    assert len(app.dataset_directories()) == 0

    dataset_directory = app.dataset_directory.new(name=dataset_directory_name_1).build()
    assert len(app.dataset_directories()) == 1
    assert len(dataset_directory.directories()) == 0

    nested_dataset_directory = (
        app.dataset_directory.new(name=dataset_directory_name_2)
        .parent(dataset_directory)
        .build()
    )
    assert len(dataset_directory.directories()) == 1
    assert len(nested_dataset_directory.datasets()) == 0

    dataset_directories = app.dataset_directories()
    assert dataset_directories[dataset_directory_name_1]
    assert dataset_directories[dataset_directory_name_2]

    dataset = (
        app.dataset.new(name=dataset_name)
        .directory(directory=nested_dataset_directory)
        .df(df1)
        .build()
    )
    nested_directory_datasets = nested_dataset_directory.datasets()
    assert len(nested_directory_datasets) == 1
    assert dataset_name in nested_directory_datasets
    assert nested_directory_datasets[dataset_name].dataset_id == dataset.dataset_id

    assert len(dataset_directory.datasets()) == 0


def test_dataset_browser_1(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name).description("Test to get dataset by name").build()
    )
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    fetched_dataset = app.datasets[dataset_name]
    assert fetched_dataset.dataset_id == dataset.dataset_id
    assert fetched_dataset.name == dataset_name


def test_dataset_browser_search_1(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name).description("Test to get dataset by name").build()
    )
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    dataset_name_substr = dataset_name.split("-", maxsplit=1)[1]
    fetched_datasets = app.datasets.search(dataset_name_substr)

    assert dataset_name in fetched_datasets
    fetched_dataset = fetched_datasets[dataset_name]

    assert fetched_dataset.dataset_id == dataset.dataset_id


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


def test_iplt_7641_datasets(
    ikigai: Ikigai,
    app_name: str,
    dataset_directory_name_1: str,
    dataset_name: str,
    df1: pd.DataFrame,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("To test that app.datasets gets all datasets")
        .build()
    )
    cleanup.callback(app.delete)

    dataset_1 = app.dataset.new(name=dataset_name).df(df1).build()

    dataset_directory = app.dataset_directory.new(name=dataset_directory_name_1).build()
    dataset_2 = (
        app.dataset.new(name=f"cloned-{dataset_name}")
        .directory(dataset_directory)
        .df(df1)
        .build()
    )

    datasets = app.datasets()
    directory_datasets = dataset_directory.datasets()
    assert datasets
    assert directory_datasets
    assert len(directory_datasets) == 1
    assert len(datasets) >= len(directory_datasets)
    assert datasets[dataset_1.name]
    assert datasets[dataset_2.name]
    with pytest.raises(KeyError):
        directory_datasets[dataset_1.name]
    assert directory_datasets[dataset_2.name]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Protocol


class ModelType(Protocol):
    @property
    def model_type(self) -> str: ...

    @property
    def sub_model_type(self) -> str: ...



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

import abc
from collections.abc import Generator


class Helpful(abc.ABC):
    """
    Abstract base class for Classes that provide help methods.
    """

    @abc.abstractmethod
    def _help(self) -> Generator[str]: ...

    def help(self) -> str:
        """
        Get formatted string of the help documentation.

        Returns
        -------
        str
            Formatted help documentation.
        """
        return "\n".join(list(self._help()))



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.ikigai import Ikigai
from ikigai.utils import (
    AppAccessLevel,
    CustomFacetAccessLevel,
    FlowStatus,
)

__all__ = [
    "AppAccessLevel",
    "CustomFacetAccessLevel",
    "FlowStatus",
    "Ikigai",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Protocol, TypedDict

from ikigai.utils import DirectoryType


class DirectoryDict(TypedDict):
    directory_id: str
    type: DirectoryType


class NamedDirectoryDict(DirectoryDict, TypedDict):
    name: str


class Directory(Protocol):
    @property
    def directory_id(self) -> str: ...

    @property
    def type(self) -> DirectoryType: ...

    def to_dict(self) -> DirectoryDict: ...



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys

# Multiple python version compatible import for deprecated decorator
if sys.version_info >= (3, 13):
    from warnings import deprecated  # noqa: F401, RUF100
else:
    from deprecated import deprecated  # noqa: F401, RUF100

# Multiple python version compatible import for typing.override
if sys.version_info >= (3, 12):
    from typing import override  # noqa: F401, RUF100
else:
    from typing_extensions import override  # noqa: F401, RUF100

# Multiple python version compatible import for Self
if sys.version_info >= (3, 11):
    from typing import Self  # noqa: F401, RUF100
else:
    from typing_extensions import Self  # noqa: F401, RUF100

# Multiple python version compatible import for StrEnum
if sys.version_info >= (3, 11):
    from enum import StrEnum  # noqa: F401, RUF100
else:
    from backports.strenum import StrEnum  # noqa: F401, RUF100

# Multiple python version compatible import for HTTPMethod
if sys.version_info >= (3, 11):
    from http import HTTPMethod
else:

    class HTTPMethod(StrEnum):
        CONNECT = "CONNECT"
        DELETE = "DELETE"
        GET = "GET"
        HEAD = "HEAD"
        OPTIONS = "OPTIONS"
        PATCH = "PATCH"
        POST = "POST"
        PUT = "PUT"
        TRACE = "TRACE"


# Multiple python version compatible import for datetime.UTC
if sys.version_info >= (3, 11):
    from datetime import UTC  # noqa: F401, RUF100
else:
    from datetime import timezone

    UTC = timezone.utc  # noqa: F401, RUF100

# Multiple python version compatible import for typing.NotRequired
if sys.version_info >= (3, 11):
    from typing import NotRequired  # noqa: F401, RUF100
else:
    from typing_extensions import NotRequired  # noqa: F401, RUF100



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from typing import Any

from ikigai.utils.missing import MISSING


def merge_dicts(obj_1: dict, obj_2: dict, *, sentinel: Any = MISSING) -> dict:
    """
    Merge two dictionaries recursively, with the second taking precedence.

    Performs a deep merge of nested dictionaries. When keys conflict between
    the two dictionaries, values from obj_2 override those from obj_1.

    Parameters
    ----------

    obj_1 : dict
        The first dictionary to merge.

    obj_2 : dict
        The second dictionary to merge (takes precedence in conflicts).

    sentinel: Any
        Sentinel value, when present in obj_2 will remove
        the corresponding key from the result.
        Avoid using None or MISSING as sentinel, instead create a sentinel
        using `object()` to ensure uniqueness.

    Returns
    -------

    dict
        The merged dictionary.

    Examples
    --------

    Simple merge with conflict resolution:

    >>> merge({'a': 1, 'b': 2}, {'b': 3, 'c': 4})
    {'a': 1, 'b': 3, 'c': 4}

    Nested dictionary merge:

    >>> merge({'a': {'b': 1}}, {'a': {'c': 2}})
    {'a': {'b': 1, 'c': 2}}

    List values are replaced, not merged:

    >>> merge({'a': {'b': [1, 2]}}, {'a': {'b': [3, 4]}})
    {'a': {'b': [3, 4]}}

    Using sentinel to remove a value from obj_1

    >>> sentinel = object()
    >>> merge({'a': 1, 'b': 2}, {'b': sentinel}, sentinel=sentinel)
    {'a': 1}
    """
    if type(obj_1) is not type(obj_2):
        return obj_2

    use_sentinel = sentinel is not MISSING
    if isinstance(obj_1, dict):
        result = dict(**obj_1)
        for k, v in obj_2.items():
            if use_sentinel and v is sentinel:
                result.pop(k, None)
                continue

            result[k] = (
                merge_dicts(result[k], v, sentinel=sentinel) if k in result else v
            )
        return result
    return obj_2



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from contextlib import ExitStack
from datetime import datetime, timedelta

import pandas as pd
import pytest

from ikigai import FlowStatus, Ikigai
from ikigai.components import Schedule


def test_flow_creation(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flows = app.flows()
    assert len(flows) == 0

    flow = app.flow.new(name=flow_name).build()

    flow_dict = flow.to_dict()
    assert flow_dict["name"] == flow_name
    assert flow_dict["schedule"] is None

    flows_after_creation = app.flows()
    assert len(flows_after_creation) == 1
    assert flows_after_creation[flow.name]

    flow.delete()
    flows_after_deletion = app.flows()
    assert len(flows_after_deletion) == 0
    with pytest.raises(KeyError):
        flows_after_deletion[flow.name]
    assert flow.flow_id not in flows_after_deletion


def test_flow_renaming(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()

    flow.rename(f"updated {flow_name}")

    flow_after_edit = app.flows().get_id(flow.flow_id)

    assert flow_after_edit.name == flow.name
    assert flow_after_edit.name == f"updated {flow_name}"


def test_flow_definition_update(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    facet_types = ikigai.facet_types

    # Not a complete flow definition, should fail when run
    initial_flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED, name=dataset.name)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .build()
    )

    flow = app.flow.new(name=flow_name).definition(initial_flow_definition).build()

    failure_log = flow.run()
    assert failure_log.status == FlowStatus.FAILED, failure_log.data

    # A complete flow definition, should succeed when run
    updated_flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED, name=dataset.name)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(
            dataset_name=f"output-{flow_name}",
            file_type="csv",
            header=True,
        )
        .build()
    )

    flow.update_definition(updated_flow_definition)

    success_log = flow.run()
    assert success_log.status == FlowStatus.SUCCESS, success_log.data


def test_flow_status(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()

    status_report = flow.status()
    assert status_report.status == FlowStatus.IDLE
    assert status_report.progress is None
    assert not status_report.message


def test_flow_clone(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    # TODO: Update test once we can nicely create pipelines
    flow = (
        app.flow.new(name=flow_name)
        .definition(
            {
                "facets": [
                    {
                        "facet_id": "input",
                        "name": dataset.name,
                        "facet_uid": "I_005",  # Imported dataset
                        "arguments": {
                            "dataset_id": dataset.dataset_id,
                            "file_type": "csv",
                            "header": True,
                            "use_raw_file": False,
                        },
                    },
                    {
                        "facet_id": "count",
                        "name": "count",
                        "facet_uid": "M_003",  # Count
                        "arguments": {
                            "output_column_name": "count",
                            "sort": True,
                            "target_columns": df1.columns.to_list()[:-2],
                        },
                    },
                    {
                        "facet_id": "output",
                        "name": "output",
                        "facet_uid": "O_005",  # Exported dataset
                        "arguments": {
                            "dataset_name": f"output-{flow_name}",
                            "file_type": "csv",
                            "header": True,
                        },
                    },
                ],
                "arrows": [
                    {
                        "arguments": {},
                        "source": "input",
                        "destination": "count",
                    },
                    {
                        "arguments": {},
                        "source": "count",
                        "destination": "output",
                    },
                ],
            }
        )
        .build()
    )

    cloned_flow = (
        app.flow.new(name=f"clone of {flow_name}").definition(definition=flow).build()
    )

    flows = app.flows()

    # TODO: Add more assert to check that the cloning did happen
    assert flows[flow.name]
    assert flows[cloned_flow.name]


def test_flow_run_success_1(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED, name=dataset.name)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.MID.PYTHON, name="count")
        .arguments(script=("import pandas as pd\ndf = data\nresult = df\n"))
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(
            dataset_name=f"output-{flow_name}",
            file_type="csv",
            header=True,
        )
        .build()
    )

    flow = app.flow.new(name=flow_name).definition(definition=flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data
    assert log.erroneous_facet_id is None, log
    assert not log.data


def test_flow_run_fail_1(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED, name=dataset.name)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.MID.PYTHON, name="failing")
        .arguments(
            script=(
                "import pandas as pd\n"
                "df = data\n"
                "raise ValueError('Expected Error')\n"
                "result = df\n"
            )
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(
            dataset_name=f"output-{flow_name}",
            file_type="csv",
            header=True,
        )
        .build()
    )

    flow = app.flow.new(name=flow_name).definition(definition=flow_definition).build()

    log = flow.run()
    assert log.status == FlowStatus.FAILED, log.data
    assert log.erroneous_facet_id, log
    failing_facets = [
        facet
        for facet in flow_definition.facets
        if facet.facet_id == log.erroneous_facet_id
    ]
    assert len(failing_facets) == 1
    assert failing_facets[0].name == "failing"
    assert log.data, log


def test_flow_run_variables(
    ikigai: Ikigai,
    app_name: str,
    dataset_name: str,
    df1: pd.DataFrame,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    dataset = app.dataset.new(name=dataset_name).df(df1).build()

    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_type=facet_types.INPUT.IMPORTED, name=dataset.name)
        .arguments(
            dataset_id=dataset.dataset_id,
            file_type="csv",
            header=True,
            use_raw_file=False,
        )
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(dataset_name=f"output-{flow_name}", file_type="csv", header=True)
        .variables(output_dataset="dataset_name")
        .build()
    )

    flow = app.flow.new(name=flow_name).definition(definition=flow_definition).build()

    run_var_output_dataset = f"var-output-{flow_name}"
    log = flow.run(output_dataset=run_var_output_dataset)
    assert log.status == FlowStatus.SUCCESS, log.data
    assert log.erroneous_facet_id is None, log
    assert not log.data

    output_dataset = app.datasets[run_var_output_dataset]
    assert output_dataset


def test_flow_directories_creation(
    ikigai: Ikigai,
    app_name: str,
    flow_directory_name_1: str,
    flow_directory_name_2: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test flow directory creation")
        .build()
    )
    cleanup.callback(app.delete)

    assert len(app.flow_directories()) == 0

    flow_directory = app.flow_directory.new(name=flow_directory_name_1).build()
    assert len(app.flow_directories()) == 1
    assert len(flow_directory.directories()) == 0

    nested_flow_directory = (
        app.flow_directory.new(name=flow_directory_name_2)
        .parent(flow_directory)
        .build()
    )
    assert len(flow_directory.directories()) == 1
    assert len(nested_flow_directory.flows()) == 0

    flow_directories = app.flow_directories()
    assert flow_directories[flow_directory_name_1]
    assert flow_directories[flow_directory_name_2]

    flow = (
        app.flow.new(name=flow_name).directory(directory=nested_flow_directory).build()
    )
    nested_directory_flows = nested_flow_directory.flows()
    assert len(nested_directory_flows) == 1
    assert flow_name in nested_directory_flows
    assert nested_directory_flows[flow_name].flow_id == flow.flow_id

    assert len(flow_directory.flows()) == 0


def test_flow_browser_1(
    ikigai: Ikigai, app_name: str, flow_name: str, cleanup: ExitStack
) -> None:
    app = ikigai.app.new(name=app_name).description("Test to get flow by name").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()

    fetched_flow = app.flows[flow_name]
    assert fetched_flow.flow_id == flow.flow_id
    assert fetched_flow.name == flow_name


def test_flow_schedule_build_1(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to build a scheduled flow with a cron string.

    This test is to ensure that the flow schedule is built correctly
    when a cron string is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to build a scheduled flow with a cron string")
        .build()
    )
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).schedule("0 0 * * *").build()

    assert flow.schedule is not None, flow.describe()

    assert flow.schedule.name == flow.name, flow.schedule
    assert flow.schedule.cron == "0 0 * * *", flow.schedule
    assert flow.schedule.start_time is not None, flow.schedule
    assert flow.schedule.end_time is None, flow.schedule


def test_flow_schedule_build_2(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to build a scheduled flow with a schedule dictionary.

    This test is to ensure that the flow schedule is built correctly
    when a schedule dictionary is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to build a scheduled flow with a schedule dictionary")
        .build()
    )
    cleanup.callback(app.delete)

    flow = (
        app.flow.new(name=flow_name)
        .schedule(
            {
                "name": flow_name,
                "cron": "0 0 * * *",
                "start_time": str(int(datetime.now().timestamp())),
            }
        )
        .build()
    )

    assert flow.schedule is not None, flow.describe()
    assert flow.schedule.name == flow.name, flow.schedule
    assert flow.schedule.cron == "0 0 * * *", flow.schedule
    assert flow.schedule.start_time is not None, flow.schedule
    assert flow.schedule.end_time is None, flow.schedule


def test_flow_schedule_update_1(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to update a scheduled flow with a cron string.

    This test is to ensure that the flow schedule is updated correctly
    when a cron string is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to update a scheduled flow with a cron string")
        .build()
    )
    cleanup.callback(app.delete)

    start_time = datetime.now()
    end_time = start_time + timedelta(days=1)
    initial_schedule = Schedule(
        name=flow_name,
        cron="0 0 * * *",
        start_time=start_time,
        end_time=end_time,
    )
    flow = app.flow.new(name=flow_name).schedule(initial_schedule).build()

    flow.update_schedule("0 */2 * * *")

    # Re-fetch the flow
    flow = app.flows[flow_name]
    assert flow.schedule is not None, flow.describe()
    assert flow.schedule.cron == "0 */2 * * *", flow.schedule
    assert flow.schedule.start_time is not None, flow.schedule
    assert flow.schedule.end_time is None, flow.schedule


def test_flow_schedule_update_2(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to update a scheduled flow with a schedule dictionary.

    This test is to ensure that the flow schedule is updated correctly
    when a schedule dictionary is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to update a scheduled flow with a schedule dictionary")
        .build()
    )
    cleanup.callback(app.delete)

    start_time = datetime.now()
    end_time = start_time + timedelta(days=1)
    initial_schedule = Schedule(
        name=flow_name,
        cron="0 0 * * *",
        start_time=start_time,
        end_time=end_time,
    )
    flow = app.flow.new(name=flow_name).schedule(initial_schedule).build()

    flow.update_schedule(
        {
            "name": flow_name,
            "cron": "0 */2 * * *",
            "start_time": str(int(datetime.now().timestamp())),
        }
    )

    # Re-fetch the flow
    flow = app.flows[flow_name]
    assert flow.schedule is not None, flow.describe()
    assert flow.schedule.cron == "0 */2 * * *", flow.schedule
    assert flow.schedule.start_time is not None, flow.schedule
    assert flow.schedule.end_time is None, flow.schedule


def test_flow_schedule_update_3(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to update a scheduled flow with a schedule object.

    This test is to ensure that the flow schedule is updated correctly
    when a schedule object is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to update a scheduled flow with a schedule object")
        .build()
    )
    cleanup.callback(app.delete)

    start_time = datetime.now()
    end_time = start_time + timedelta(weeks=1)
    initial_schedule = Schedule(
        name=flow_name,
        cron="0 0 * * *",
        start_time=start_time,
        end_time=end_time,
    )
    flow = app.flow.new(name=flow_name).schedule(initial_schedule).build()

    new_schedule = Schedule(
        name=flow_name,
        cron="0 */2 * * *",
        start_time=start_time,
        end_time=end_time,
    )
    flow.update_schedule(new_schedule)

    # Re-fetch the flow
    flow = app.flows[flow_name]
    assert flow.schedule is not None, flow.describe()
    schedule = flow.schedule
    assert schedule.cron == new_schedule.cron, schedule
    assert schedule.start_time == new_schedule.start_time, schedule.start_time
    assert schedule.end_time == new_schedule.end_time, schedule.end_time


def test_flow_schedule_build_3(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to build a scheduled flow with a schedule object.

    This test is to ensure that the flow schedule is built correctly
    when a schedule object is provided.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to build a scheduled flow with a schedule object")
        .build()
    )
    cleanup.callback(app.delete)

    schedule = Schedule(
        name=flow_name,
        cron="0 0 * * *",
        start_time=datetime.now(),
        end_time=None,
    )
    flow = app.flow.new(name=flow_name).schedule(schedule).build()

    assert flow.schedule is not None, flow.describe()

    assert flow.schedule.name == flow.name, flow.schedule
    assert flow.schedule.cron == "0 0 * * *", flow.schedule
    assert flow.schedule.start_time is not None, flow.schedule
    assert flow.schedule.end_time is None, flow.schedule


def test_flow_schedule_update_4(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    """
    Test to remove the schedule from a flow.

    This test is to ensure that the flow schedule is removed correctly.
    """
    app = (
        ikigai.app.new(name=app_name)
        .description("Test to remove the schedule from a flow")
        .build()
    )
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).schedule("0 0 * * *").build()

    flow.update_schedule(None)

    # Re-fetch the flow
    flow = app.flows[flow_name]
    assert flow.schedule is None, flow.describe()


def test_flow_browser_search_1(
    ikigai: Ikigai, app_name: str, flow_name: str, cleanup: ExitStack
) -> None:
    app = ikigai.app.new(name=app_name).description("Test to get flow by name").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()

    flow_name_substr = flow_name.split("-", maxsplit=1)[1]
    fetched_flows = app.flows.search(flow_name_substr)

    assert flow_name in fetched_flows
    fetched_flow = fetched_flows[flow_name]

    assert fetched_flow.flow_id == flow.flow_id


def test_flow_high_volume_preference(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).high_volume_preference(optimize=True).build()

    # Verify that the high volume preference is set correctly
    flow_details = flow.describe()
    assert flow_details["high_volume_preference"] is True, flow_details


def test_flow_high_volume_preference_update(
    ikigai: Ikigai,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("App to test flow run").build()
    cleanup.callback(app.delete)

    facet_types = ikigai.facet_types
    flow_definition = (
        ikigai.builder.facet(facet_types.MID.Python)
        .arguments(
            script=(
                "import pandas as pd\n"
                "df = data\n"
                "raise ValueError('Expected Error')\n"
                "result = df\n"
            )
        )
        .build()
    )

    flow = (
        app.flow.new(name=flow_name)
        .definition(flow_definition)
        .high_volume_preference(optimize=False)
        .build()
    )

    flow_details = flow.describe()
    assert flow_details["high_volume_preference"] is False, flow_details
    assert flow_details["definition"]["facets"], flow_details["definition"]

    # Update the high volume preference
    flow.update_high_volume_preference(optimize=True)

    updated_flow_details = flow.describe()
    assert updated_flow_details["high_volume_preference"] is True, updated_flow_details
    assert updated_flow_details["definition"]["facets"], updated_flow_details[
        "definition"
    ]
    assert updated_flow_details["definition"] == flow_details["definition"]

    # Check that high volume preference persists after updating the definition
    new_flow_definition = ikigai.builder.build()
    flow.update_definition(definition=new_flow_definition)
    final_flow_details = flow.describe()

    assert final_flow_details["high_volume_preference"] is True, final_flow_details


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


def test_iplt_7641_flows(
    ikigai: Ikigai,
    app_name: str,
    flow_directory_name_1: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("An app to test that app.flows gets all flows")
        .build()
    )
    cleanup.callback(app.delete)

    flow = app.flow.new(name=flow_name).build()

    flow_directory = app.flow_directory.new(name=flow_directory_name_1).build()
    cloned_flow = (
        app.flow.new(name=f"cloned-{flow_name}")
        .directory(flow_directory)
        .definition(flow)
        .build()
    )

    flows = app.flows()
    directory_flows = flow_directory.flows()
    assert flows
    assert directory_flows
    assert len(directory_flows) == 1
    assert len(flows) >= len(directory_flows)
    assert flows[flow.name]
    assert flows[cloned_flow.name]
    with pytest.raises(KeyError):
        directory_flows[flow.name]
    assert directory_flows[cloned_flow.name]



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

from typing import Any

from ikigai.utils.compatibility import Self, StrEnum

# -------------------------------------------------------------------------------------
# App Related Enums


class AppAccessLevel(StrEnum):
    OWNER = "OWNER"
    BUILDER = "BUILDER"
    VIEWER = "VIEWER"


# -------------------------------------------------------------------------------------
# Custom Facet Related Enums


class CustomFacetAccessLevel(StrEnum):
    ADMIN = "ADMIN"
    EDITOR = "EDIT"
    VIEWER = "VIEW"
    RUNNER = "RUN"


class CustomFacetArgumentType(StrEnum):
    BOOLEAN = "bool"
    INTEGER = "int"
    STRING = "str"
    FLOAT = "float"

    @classmethod
    def from_value(cls, value: Any) -> Self:
        type_name = type(value).__name__
        return cls(type_name)

    @property
    def python_type(self) -> type[bool | int | str | float]:
        return {
            CustomFacetArgumentType.BOOLEAN: bool,
            CustomFacetArgumentType.INTEGER: int,
            CustomFacetArgumentType.STRING: str,
            CustomFacetArgumentType.FLOAT: float,
        }[self]

    def to_facet_argument_type(self) -> FacetArgumentType:
        return {
            CustomFacetArgumentType.BOOLEAN: FacetArgumentType.BOOLEAN,
            CustomFacetArgumentType.INTEGER: FacetArgumentType.NUMBER,
            CustomFacetArgumentType.STRING: FacetArgumentType.TEXT,
            CustomFacetArgumentType.FLOAT: FacetArgumentType.NUMBER,
        }[self]


# -------------------------------------------------------------------------------------
# Dataset Related Enums


class DatasetDownloadStatus(StrEnum):
    SUCCESS = "SUCCESS"
    IN_PROGRESS = "IN_PROGRESS"
    FAILED = "FAILED"


class DatasetDataType(StrEnum):
    NUMERIC = "NUMERIC"
    TEXT = "TEXT"
    CATEGORICAL = "CATEGORICAL"
    TIME = "TIME"


# -------------------------------------------------------------------------------------
# Directory Related Enums


class DirectoryType(StrEnum):
    APP = "PROJECT"
    DATASET = "DATASET"
    FLOW = "PIPELINE"
    MODEL = "MODEL"


# -------------------------------------------------------------------------------------
# Flow Related Enums


class FlowStatus(StrEnum):
    SCHEDULED = "SCHEDULED"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    FAILED = "FAILED"
    IDLE = "IDLE"
    UNKNOWN = "UNKNOWN"
    SUCCESS = "SUCCESS"  # Not available via /component/is-pipeline-running

    def __repr__(self) -> str:
        return self.value


# -------------------------------------------------------------------------------------
# Specs Related Enums


class FacetArgumentType(StrEnum):
    MAP = "MAP"
    BOOLEAN = "BOOLEAN"
    TEXT = "TEXT"
    NUMBER = "NUMBER"


class ModelHyperparameterType(StrEnum):
    MAP = "MAP"
    BOOLEAN = "BOOLEAN"
    TEXT = "TEXT"
    NUMBER = "NUMBER"


class ModelParameterType(StrEnum):
    BOOLEAN = "BOOLEAN"
    TEXT = "TEXT"
    NUMBER = "NUMBER"


# -------------------------------------------------------------------------------------
